#pragma once

#include <map>
#include <string>

#include "ProviderSegmentBase.h"
#include "detail/MaxHistorySize.h"
#include "detail/TypedEntityContainerBase.h"


namespace armarx::armem::base
{

    /**
     * @brief Data of a core segment containing multiple provider segments.
     */
    template <class _ProviderSegmentT, class _Derived>
    class CoreSegmentBase :
        virtual public detail::TypedEntityContainerBase<_ProviderSegmentT, typename _ProviderSegmentT::EntityT, _Derived>,
        virtual public detail::MaxHistorySize
    {
        using Base = detail::TypedEntityContainerBase<_ProviderSegmentT, typename _ProviderSegmentT::EntityT, _Derived>;

    public:

        using typename Base::DerivedT;
        using typename Base::ContainerT;

        using ProviderSegmentT = _ProviderSegmentT;
        using EntityT = typename ProviderSegmentT::EntityT;
        using EntitySnapshotT = typename EntityT::EntitySnapshotT;
        using EntityInstanceT = typename EntitySnapshotT::EntityInstanceT;


    public:

        CoreSegmentBase& operator=(const CoreSegmentBase& other)
        {
            other._copySelf(*this);
            return *this;
        }

        inline const std::string& name() const
        {
            return this->id().coreSegmentName;
        }
        inline std::string& name()
        {
            return const_cast<std::string&>(const_cast<const CoreSegmentBase*>(this)->name());
        }


        inline const std::map<std::string, ProviderSegmentT>& providerSegments() const
        {
            return _container;
        }
        inline std::map<std::string, ProviderSegmentT>& providerSegments()
        {
            return const_cast<std::map<std::string, ProviderSegmentT>&>(const_cast<const CoreSegmentBase*>(this)->providerSegments());
        }


        bool hasProviderSegment(const std::string& name) const
        {
            return _container.count(name) > 0;
        }

        ProviderSegmentT& getProviderSegment(const std::string& name)
        {
            return const_cast<ProviderSegmentT&>(const_cast<const CoreSegmentBase*>(this)->getProviderSegment(name));
        }

        const ProviderSegmentT& getProviderSegment(const std::string& name) const
        {
            auto it = this->_container.find(name);
            if (it != _container.end())
            {
                return it->second;
            }
            else
            {
                throw armem::error::MissingEntry("provider segment", name, getLevelName(), this->getKeyString());
            }
        }

        using Base::_checkContainerName;
        const EntityT& getEntity(const MemoryID& id) const override
        {
            _checkContainerName(id.coreSegmentName, this->getKeyString());
            return getProviderSegment(id.providerSegmentName).getEntity(id);
        }

        virtual MemoryID update(const EntityUpdate& update) override
        {
            _checkContainerName(update.entityID.coreSegmentName, this->name());

            auto it = this->_container.find(update.entityID.providerSegmentName);
            if (it != _container.end())
            {
                return it->second.update(update);
            }
            else
            {
                throw armem::error::MissingEntry("provider segment", update.entityID.providerSegmentName, getLevelName(), this->getKeyString());
            }
        }

        /**
         * @brief Add an empty provider segment with the given name and optional provider segment type.
         * @param name The segment name.
         * @param providerSegmentType The provider type. If nullptr, the core segment type is used.
         * @return The added provider segment.
         */
        ProviderSegmentT& addProviderSegment(const std::string& name, aron::typenavigator::ObjectNavigatorPtr providerSegmentType = nullptr)
        {
            aron::typenavigator::ObjectNavigatorPtr type = providerSegmentType ? providerSegmentType : this->aronType();
            return addProviderSegment(ProviderSegmentT(name, type));
        }

        /// Copy and insert a provider segment.
        ProviderSegmentT& addProviderSegment(const ProviderSegmentT& providerSegment)
        {
            return addProviderSegment(ProviderSegment(providerSegment));
        }

        /// Move and insert a provider segment.
        ProviderSegmentT& addProviderSegment(ProviderSegmentT&& providerSegment)
        {
            if (hasProviderSegment(providerSegment.name()))
            {
                throw armem::error::ContainerEntryAlreadyExists(
                    providerSegment.getLevelName(), providerSegment.name(), getLevelName(), this->getKeyString());
            }

            auto it = _container.emplace(providerSegment.name(), std::move(providerSegment)).first;
            it->second.id().setCoreSegmentID(this->id());
            it->second.setMaxHistorySize(_maxHistorySize);
            return it->second;
        }


        /**
         * @brief Sets the maximum history size of entities in this segment.
         * This affects all current entities as well as new ones.
         * @see Entity::setMaxHistorySize()
         */
        void setMaxHistorySize(long maxSize) override
        {
            MaxHistorySize::setMaxHistorySize(maxSize);
            for (auto& [name, seg] : _container)
            {
                seg.setMaxHistorySize(maxSize);
            }
        }

        virtual bool equalsDeep(const CoreSegmentBase& other) const
        {
            //std::cout << "CoreSegment::equalsDeep" << std::endl;
            if (this->size() != other.size())
            {
                return false;
            }
            for (const auto& [key, provider] : _container)
            {
                if (not other.hasProviderSegment(key))
                {
                    return false;
                }
                if (not provider.equalsDeep(other.getProviderSegment(key)))
                {
                    return false;
                }
            }
            return true;
        }

        std::string getLevelName() const override
        {
            return "core segment";
        }

        std::string getKeyString() const override
        {
            return this->name();
        }


    protected:

        using Base::_container;

    };

}
