#pragma once

#include <memory>
#include <vector>

#include "../MemoryID.h"
#include "../Time.h"

#include "EntityInstanceBase.h"
#include "detail/MemoryContainerBase.h"


namespace armarx::armem::base
{
    /**
     * @brief Data of an entity at one point in time.
     */
    template <class _EntityInstanceT, class _Derived>
    class EntitySnapshotBase :
        virtual public detail::MemoryContainerBase<std::vector<_EntityInstanceT>, _Derived>
    {
        using Base = detail::MemoryContainerBase<std::vector<_EntityInstanceT>, _Derived>;

    public:

        using typename Base::DerivedT;
        using typename Base::ContainerT;

        using EntityInstanceT = _EntityInstanceT;


    public:

        EntitySnapshotBase& operator=(const EntitySnapshotBase& other)
        {
            other._copySelf(*this);
            return *this;
        }

        virtual bool equalsDeep(const EntitySnapshotBase& other) const
        {
            //std::cout << "EntitySnapshot::equalsDeep" << std::endl;
            if (this->size() != other.size())
            {
                return false;
            }
            int i = 0;
            for (const auto& instance : _container)
            {
                if (not instance.equalsDeep(other.getInstance(i)))
                {
                    return false;
                }
                i++;
            }
            return true;
        }

        inline Time& time()
        {
            return this->id().timestamp;
        }

        inline const Time& time() const
        {
            return this->id().timestamp;
        }


        inline const std::vector<EntityInstanceT>& instances() const
        {
            return _container;
        }
        inline std::vector<EntityInstanceT>& instances()
        {
            return const_cast<std::vector<EntityInstanceT>&>(const_cast<const EntitySnapshotBase*>(this)->instances());
        }


        void update(const EntityUpdate& update, std::optional<MemoryID> parentID = std::nullopt)
        {
            if (parentID)
            {
                this->id() = *parentID;
            }
            time() = update.timeCreated;

            _container.clear();
            for (int i = 0; i < int(update.instancesData.size()); ++i)
            {
                EntityInstanceT& data = _container.emplace_back(i, this->id());
                data.update(update, i);
            }
        }


        bool hasInstance(int index) const
        {
            size_t si = size_t(index);
            return index >= 0 && si < _container.size();
        }

        /**
         * @brief Get the given instance.
         * @param index The instance's index.
         * @return The instance.
         * @throw `armem::error::MissingEntry` If the given index is invalid.
         */
        EntityInstanceT& getInstance(int index)
        {
            return const_cast<EntityInstanceT&>(const_cast<const EntitySnapshotBase*>(this)->getInstance(index));
        }

        const EntityInstanceT& getInstance(int index) const
        {
            if (hasInstance(index))
            {
                return _container[static_cast<size_t>(index)];
            }
            else
            {
                throw armem::error::MissingEntry(EntityInstanceT().getLevelName(), std::to_string(index), getLevelName(), toDateTimeMilliSeconds(time()));
            }
        }

        /**
         * @brief Get the given instance.
         * @param index The instance's index.
         * @return The instance.
         * @throw `armem::error::MissingEntry` If the given index is invalid.
         * @throw `armem::error::InvalidMemoryID` If memory ID does not have an instance index.
         */
        EntityInstanceT& getInstance(const MemoryID& id)
        {
            return const_cast<EntityInstanceT&>(const_cast<const EntitySnapshotBase*>(this)->getInstance(id));
        }

        const EntityInstanceT& getInstance(const MemoryID& id) const
        {
            if (!id.hasInstanceIndex())
            {
                throw armem::error::InvalidMemoryID(id, "ID has no instance index.");
            }
            return getInstance(id.instanceIndex);
        }


        /**
         * @brief Add a single instance with data.
         * @param instance The instance.
         * @return The stored instance.
         * @throw `armem::error::InvalidArgument` If the given index is invalid. Must be equal to container.size() or -1 (meaning push_back)
         */
        EntityInstanceT& addInstance(const EntityInstanceT& instance)
        {
            return addInstance(EntityInstanceT(instance));
        }

        EntityInstanceT& addInstance(EntityInstanceT&& instance)
        {
            if (instance.index() > 0 && static_cast<size_t>(instance.index()) < _container.size())
            {
                throw error::InvalidArgument(std::to_string(instance.index()), "EntitySnapshot::addInstance",
                                             "Cannot add an EntityInstance because its index already exists.");
            }
            if (instance.index() > 0 && static_cast<size_t>(instance.index()) > _container.size())
            {
                throw error::InvalidArgument(std::to_string(instance.index()), "EntitySnapshot::addInstance",
                                             "Cannot add an EntityInstance because its index is too big.");
            }

            int new_index = _container.size();
            auto& it = _container.emplace_back(std::move(instance));
            it.index() = new_index;
            return it;
        }

        std::string getKeyString() const override
        {
            return toDateTimeMilliSeconds(this->time());
        }

        std::string getLevelName() const override
        {
            return "entity snapshot";
        }


    protected:

        using Base::_container;

    };

}
