#pragma once

#include <map>
#include <string>

#include <ArmarXCore/core/logging/Logging.h>

#include "CoreSegmentBase.h"
#include "detail/EntityContainerBase.h"


namespace armarx::armem::base
{

    /**
     * @brief Data of a memory consisting of multiple core segments.
     */
    template <class _CoreSegmentT, class _Derived>
    class MemoryBase :
        virtual public detail::EntityContainerBase<_CoreSegmentT, typename _CoreSegmentT::ProviderSegmentT::EntityT, _Derived>
    {
        using Base = detail::EntityContainerBase<_CoreSegmentT, typename _CoreSegmentT::ProviderSegmentT::EntityT, _Derived>;

    public:

        using typename Base::DerivedT;
        using typename Base::ContainerT;

        using CoreSegmentT = _CoreSegmentT;
        using ProviderSegmentT = typename CoreSegmentT::ProviderSegmentT;
        using EntityT = typename ProviderSegmentT::EntityT;
        using EntitySnapshotT = typename EntityT::EntitySnapshotT;
        using EntityInstanceT = typename EntitySnapshotT::EntityInstanceT;


    public:

        MemoryBase& operator=(const MemoryBase& other)
        {
            other._copySelf(*this);
            return *this;
        }


        inline const std::string& name() const
        {
            return this->id().memoryName;
        }
        inline std::string& name()
        {
            return const_cast<std::string&>(const_cast<const MemoryBase*>(this)->name());
        }


        inline const std::map<std::string, CoreSegmentT>& coreSegments() const
        {
            return _container;
        }
        inline std::map<std::string, CoreSegmentT>& coreSegments()
        {
            return const_cast<std::map<std::string, CoreSegmentT>&>(const_cast<const MemoryBase*>(this)->coreSegments());
        }


        bool hasCoreSegment(const std::string& name) const
        {
            return _container.count(name) > 0;
        }

        CoreSegmentT& getCoreSegment(const std::string& name)
        {
            return const_cast<CoreSegmentT&>(const_cast<const MemoryBase*>(this)->getCoreSegment(name));
        }

        const CoreSegmentT& getCoreSegment(const std::string& name) const
        {
            auto it = this->_container.find(name);
            if (it != _container.end())
            {
                return it->second;
            }
            else
            {
                throw armem::error::MissingEntry(CoreSegmentT().getLevelName(), name, getLevelName(), this->name());
            }
        }

        using Base::_checkContainerName;
        const EntityT& getEntity(const MemoryID& id) const override
        {
            _checkContainerName(id.memoryName, this->name());
            return getCoreSegment(id.coreSegmentName).getEntity(id);
        }

        /**
         * @brief Add an empty core segment with the given name.
         * @param name The core segment name.
         * @param coreSegmentType The core segment type (optional).
         * @return The added core segment.
         */
        CoreSegmentT& addCoreSegment(const std::string& name, aron::typenavigator::ObjectNavigatorPtr coreSegmentType = nullptr)
        {
            return addCoreSegment(CoreSegmentT(name, coreSegmentType));
        }
        /// Copy and insert a core segment.
        CoreSegmentT& addCoreSegment(const CoreSegmentT& coreSegment)
        {
            return addCoreSegment(CoreSegmentT(coreSegment));
        }
        /// Move and insert a core segment.
        CoreSegmentT& addCoreSegment(CoreSegmentT&& coreSegment)
        {
            if (_container.count(coreSegment.name()) > 0)
            {
                throw armem::error::ContainerEntryAlreadyExists(coreSegment.getLevelName(), coreSegment.name(),
                        this->getLevelName(), this->name());
            }
            auto it = _container.emplace(coreSegment.name(), std::move(coreSegment)).first;
            it->second.id().setMemoryID(this->id());
            return it->second;
        }

        /**
         * @brief Add multiple core segments.
         * @param The core segment names.
         * @return The core segments. The contained pointers are never null.
         */
        std::vector<CoreSegmentT*> addCoreSegments(const std::vector<std::string>& names)
        {
            std::vector<CoreSegmentT*> segments;
            for (const auto& name : names)
            {
                try
                {
                    segments.push_back(&addCoreSegment(name));
                }
                catch (const armem::error::ContainerEntryAlreadyExists& e)
                {
                    ARMARX_INFO << e.what() << "\nIgnoring multiple addition.";
                }
            }
            return segments;
        }

        virtual MemoryID update(const EntityUpdate& update) override
        {
            _checkContainerName(update.entityID.memoryName, this->name());

            auto it = _container.find(update.entityID.coreSegmentName);
            if (it != _container.end())
            {
                return it->second.update(update);
            }
            else
            {
                throw armem::error::MissingEntry(CoreSegmentT().getLevelName(), update.entityID.coreSegmentName, getLevelName(), this->name());
            }
        }

        virtual bool equalsDeep(const MemoryBase& other) const
        {
            //std::cout << "Memory::equalsDeep" << std::endl;
            if (this->size() != other.size())
            {
                return false;
            }
            for (const auto& [key, core] : _container)
            {
                if (not other.hasCoreSegment(key))
                {
                    return false;
                }
                if (not core.equalsDeep(other.getCoreSegment(key)))
                {
                    return false;
                }
            }
            return true;
        }

        std::string getLevelName() const override
        {
            return "memory";
        }

        std::string getKeyString() const override
        {
            return this->name();
        }


    protected:

        using Base::_container;

    };
}
