#pragma once

#include "../../Commit.h"
#include "../../error/ArMemError.h"

#include "MemoryContainerBase.h"

#include "../EntityBase.h"
#include "../EntitySnapshotBase.h"


namespace armarx::armem::base::detail
{

    /**
     * @brief A container of entities at some point in the hierarchy.
     *
     * Can be updated by multiple entity updates.
     */
    template <class _ValueT, class _EntityT, class _Derived>
    class EntityContainerBase :
        virtual public MemoryContainerBase<std::map<std::string, _ValueT>, _Derived>
    {
        using Base = MemoryContainerBase<std::map<std::string, _ValueT>, _Derived>;

    public:

        using DerivedT = _Derived;
        using ValueT = _ValueT;

        using EntityT = _EntityT;
        using EntitySnapshotT = typename EntityT::EntitySnapshotT;
        using EntityInstanceT = typename EntitySnapshotT::EntityInstanceT;


    public:

        EntityContainerBase& operator=(const EntityContainerBase& other)
        {
            other._copySelf(*this);
            return *this;
        }

        /**
         * @brief Store all updates in `commit`.
         * @param commit The commit.
         * @return The resulting memory IDs.
         */
        std::vector<MemoryID> update(const Commit& commit)
        {
            std::vector<MemoryID> ids;
            for (const auto& update : commit.updates)
            {
                ids.push_back(this->update(update));
            }
            return ids;
        }
        /**
         * @brief Store the given update.
         * @param update The update.
         * @return The resulting entity snapshot's ID.
         */
        virtual MemoryID update(const EntityUpdate& update) = 0;


        /**
         * @brief Retrieve an entity.
         * @param id The entity ID.
         * @return The entity.
         * @throw An exception deriving from `armem::error::ArMemError` if the entity is missing.
         */
        virtual _EntityT& getEntity(const MemoryID& id)
        {
            return const_cast<_EntityT&>(const_cast<const EntityContainerBase*>(this)->getEntity(id));
        }
        virtual const _EntityT& getEntity(const MemoryID& id) const = 0;

        /**
         * @brief Retrieve an entity snapshot.
         *
         * Uses `getEntity()` to retrieve the respective entity.
         *
         * @param id The snapshot ID.
         * @return The entity snapshot.
         * @throw An exception deriving from `armem::error::ArMemError` if the snapshot is missing.
         */
        virtual EntitySnapshotT& getEntitySnapshot(const MemoryID& id)
        {
            return const_cast<EntitySnapshotT&>(const_cast<const EntityContainerBase*>(this)->getEntitySnapshot(id));
        }

        virtual const EntitySnapshotT& getEntitySnapshot(const MemoryID& id) const
        {
            const _EntityT& entity = getEntity(id);

            if (id.hasTimestamp())
            {
                return entity.getSnapshot(id);
            }
            else
            {
                return entity.getLatestSnapshot();
            }
        }

        virtual EntityInstanceT& getEntityInstance(const MemoryID& id)
        {
            return getEntitySnapshot(id).getInstance(id);
        }

        virtual const EntityInstanceT& getEntityInstance(const MemoryID& id) const
        {
            return getEntitySnapshot(id).getInstance(id);
        }
    };

}
