#include "TreeWidget.h"

#include <RobotAPI/libraries/armem_gui/instance/sanitize_typename.h>

#include <RobotAPI/libraries/aron/core/navigator/type/container/Object.h>

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include <SimoxUtility/algorithm/string.h>

#include <QHeaderView>


namespace armarx::armem::gui::memory
{

    TreeWidget::TreeWidget()
    {
        initBuilders();
        initWidget();
    }

    void TreeWidget::initWidget()
    {
        clear();
        QStringList columns;
        columns.insert(int(Columns::KEY), "Key");
        columns.insert(int(Columns::SIZE), "#");
        columns.insert(int(Columns::TYPE), "Type");
        columns.insert(int(Columns::LEVEL), "Level");
        columns.insert(int(Columns::ID), "Memory ID");
        setColumnCount(columns.size());
        setHeaderLabels(columns);

        header()->setMinimumSectionSize(25);
        header()->resizeSection(int(Columns::KEY), 250);
        header()->resizeSection(int(Columns::SIZE), 25);
        header()->setTextElideMode(Qt::TextElideMode::ElideRight);


        connect(this, &This::updated, this, &This::handleSelection);
        connect(this, &QTreeWidget::currentItemChanged, this, &This::handleSelection);

        connect(this, &This::memorySelected, this, &This::itemSelected);
        connect(this, &This::coreSegmentSelected, this, &This::itemSelected);
        connect(this, &This::providerSegmentSelected, this, &This::itemSelected);
        connect(this, &This::entitySelected, this, &This::itemSelected);
        connect(this, &This::snapshotSelected, this, &This::itemSelected);
        connect(this, &This::instanceSelected, this, &This::itemSelected);
    }

    void TreeWidget::initBuilders()
    {
        workingmemoryBuilder.setExpand(true);
        workingmemoryBuilder.setMakeItemFn([this](const std::string & name, const wm::Memory * memory)
        {
            return makeItem(name, *memory);
        });
        workingmemoryBuilder.setUpdateItemFn([this](const std::string&, const wm::Memory * memory, QTreeWidgetItem * memoryItem)
        {
            updateContainerItem(*memory, memoryItem);
            if (memoryItem)
            {
                updateChildren(*memory, memoryItem);
            }
            return true;
        });


        workingmemoryCoreSegmentBuilder.setExpand(true);
        workingmemoryCoreSegmentBuilder.setMakeItemFn([this](const std::string & name, const wm::CoreSegment & coreSeg)
        {
            return makeItem(name, coreSeg);
        });
        workingmemoryCoreSegmentBuilder.setUpdateItemFn([this](const std::string&, const wm::CoreSegment & coreSeg, QTreeWidgetItem * coreSegItem)
        {
            updateTypedContainerItem(coreSeg, coreSegItem);
            updateChildren(coreSeg, coreSegItem);
            return true;
        });

        workingmemoryProvSegmentBuilder.setExpand(true);
        workingmemoryProvSegmentBuilder.setMakeItemFn([this](const std::string & name, const wm::ProviderSegment & provSeg)
        {
            return makeItem(name, provSeg);
        });
        workingmemoryProvSegmentBuilder.setUpdateItemFn([this](const std::string&, const wm::ProviderSegment & provSeg, QTreeWidgetItem * provSegItem)
        {
            updateTypedContainerItem(provSeg, provSegItem);
            updateChildren(provSeg, provSegItem);
            return true;
        });

        // entityBuilder.setExpand(true);
        workingmemoryEntityBuilder.setMakeItemFn([this](const std::string & name, const wm::Entity & entity)
        {
            return makeItem(name, entity);
        });
        workingmemoryEntityBuilder.setUpdateItemFn([this](const std::string&, const wm::Entity & entity, QTreeWidgetItem *  entityItem)
        {
            updateContainerItem(entity, entityItem);
            updateChildren(entity, entityItem);
            return true;
        });

        workingmemorySnapshotBuilder.setMakeItemFn([this](const armem::Time & time, const wm::EntitySnapshot & snapshot)
        {
            QTreeWidgetItem* item = makeItem(toDateTimeMilliSeconds(time), snapshot);
            item->setData(int(Columns::KEY), Qt::ItemDataRole::UserRole, QVariant(static_cast<qlonglong>(snapshot.time().toMicroSeconds())));
            return item;
        });
        workingmemorySnapshotBuilder.setCompareFn([](const auto & pair, QTreeWidgetItem * item)
        {
            const wm::EntitySnapshot& snapshot = pair.second;
            return armarx::detail::compare(static_cast<qlonglong>(snapshot.time().toMicroSeconds()),
                                           item->data(int(Columns::KEY), Qt::ItemDataRole::UserRole).toLongLong());
        });
        workingmemorySnapshotBuilder.setUpdateItemFn([this](const armem::Time&, const wm::EntitySnapshot & snapshot, QTreeWidgetItem * snapshotItem)
        {
            updateContainerItem(snapshot, snapshotItem);
            updateChildren(snapshot, snapshotItem);
            return true;
        });

        workingmemoryInstanceBuilder.setMakeItemFn([this](const wm::EntityInstance & instance)
        {
            QTreeWidgetItem* item = makeItem("", instance);
            return item;
        });
        workingmemoryInstanceBuilder.setCompareFn([](const wm::EntityInstance & lhs, QTreeWidgetItem * rhsItem)
        {
            return armarx::detail::compare(lhs.index(), rhsItem->text(0).toInt());
        });
        workingmemoryInstanceBuilder.setUpdateItemFn([this](const wm::EntityInstance & instance, QTreeWidgetItem * instanceItem)
        {
            updateItemItem(instance, instanceItem);
            updateChildren(instance, instanceItem);
            return true;
        });

    }


    void TreeWidget::update(const armem::wm::Memory& memory)
    {
        // Removing elements during the update can create unwanted signals triggering selection handling.
        handleSelections = false;
        updateChildren(memory, this);
        handleSelections = true;
        emit updated();
    }

    void TreeWidget::update(const std::map<std::string, const armem::wm::Memory*>& memories)
    {
        handleSelections = false;
        updateChildren(memories, this);
        handleSelections = true;
        emit updated();
    }

    std::optional<MemoryID> TreeWidget::selectedID() const
    {
        return _selectedID;
    }

    void TreeWidget::handleSelection()
    {
        if (!handleSelections)
        {
            return;
        }
        QTreeWidgetItem* item = this->currentItem();
        if (!item)
        {
            return;
        }

        MemoryID id(item->data(int(Columns::ID), Qt::UserRole).toString().toStdString());

        // Has the selection changed due to an update? If yes, signal this.
        if (!_selectedID || id != *_selectedID)
        {
            _selectedID = id;

            const std::string levelName = item->data(int(Columns::LEVEL), Qt::UserRole).toString().toStdString();
            if (levelName == wm::Memory().getLevelName())
            {
                emit memorySelected(*_selectedID);
            }
            else if (levelName == wm::CoreSegment().getLevelName())
            {
                emit coreSegmentSelected(*_selectedID);
            }
            else if (levelName == wm::ProviderSegment().getLevelName())
            {
                emit providerSegmentSelected(*_selectedID);
            }
            else if (levelName == wm::Entity().getLevelName())
            {
                emit entitySelected(*_selectedID);
            }
            else if (levelName == wm::EntitySnapshot().getLevelName())
            {
                emit snapshotSelected(*_selectedID);
            }
            else if (levelName == wm::EntityInstance().getLevelName())
            {
                emit instanceSelected(*_selectedID);
            }
        }

        // Either selection or its contents have changed.
        emit selectedItemChanged(*_selectedID);
    }


    void TreeWidget::updateChildren(const armem::wm::Memory& memory, QTreeWidget* tree)
    {
        updateChildren(std::map<std::string, const armem::wm::Memory*> {{memory.name(), &memory}}, tree);
    }

    void TreeWidget::updateChildren(const std::map<std::string, const armem::wm::Memory*>& memories, QTreeWidget* tree)
    {
        workingmemoryBuilder.updateTree(tree, memories);
    }


    void TreeWidget::updateChildren(const armem::wm::Memory& memory, QTreeWidgetItem* memoryItem)
    {
        workingmemoryCoreSegmentBuilder.updateTree(memoryItem, memory.coreSegments());
    }

    void TreeWidget::updateChildren(const armem::wm::CoreSegment& coreSeg, QTreeWidgetItem* coreSegItem)
    {
        workingmemoryProvSegmentBuilder.updateTree(coreSegItem, coreSeg.providerSegments());
    }

    void TreeWidget::updateChildren(const armem::wm::ProviderSegment& provSeg, QTreeWidgetItem* provSegItem)
    {
        workingmemoryEntityBuilder.updateTree(provSegItem, provSeg.entities());
    }

    void TreeWidget::updateChildren(const armem::wm::Entity& entity, QTreeWidgetItem* entityItem)
    {
        workingmemorySnapshotBuilder.updateTree(entityItem, entity.history());
    }

    void TreeWidget::updateChildren(const armem::wm::EntitySnapshot& snapshot, QTreeWidgetItem* snapshotItem)
    {
        workingmemoryInstanceBuilder.updateTree(snapshotItem, snapshot.instances());
    }

    void TreeWidget::updateChildren(const armem::wm::EntityInstance& data, QTreeWidgetItem* dataItem)
    {
        (void) data, (void) dataItem;
    }

    QTreeWidgetItem* TreeWidget::makeItem(const std::string& key, const armem::base::detail::MemoryItem& memoryItem)
    {
        (void) key;
        QStringList columns;
        columns.insert(int(Columns::KEY), QString::fromStdString(memoryItem.getKeyString()));
        columns.insert(int(Columns::SIZE), "");
        columns.insert(int(Columns::TYPE), "");
        columns.insert(int(Columns::LEVEL), QString::fromStdString(simox::alg::capitalize_words(memoryItem.getLevelName())));
        columns.insert(int(Columns::ID), QString::fromStdString(memoryItem.id().str()));

        QTreeWidgetItem* item = new QTreeWidgetItem(columns);
        item->setData(int(Columns::LEVEL), Qt::UserRole, QString::fromStdString(memoryItem.getLevelName()));
        item->setData(int(Columns::ID), Qt::UserRole, QString::fromStdString(memoryItem.id().str()));
        item->setTextAlignment(int(Columns::SIZE), Qt::AlignRight);
        return item;
    }

    void TreeWidget::updateItemItem(const armem::base::detail::MemoryItem& level, QTreeWidgetItem* item)
    {
        (void) level, (void) item;
    }

    template <class... T>
    void TreeWidget::updateContainerItem(
        const armem::wm::detail::MemoryContainer<T...>& container, QTreeWidgetItem* item)
    {
        updateItemItem(container, item);
        item->setText(int(Columns::SIZE), QString::number(container.size()));
    }

    template <class... T>
    void TreeWidget::updateTypedContainerItem(
        const armem::wm::detail::TypedEntityContainer<T...>& container, QTreeWidgetItem* item)
    {
        updateContainerItem(container, item);
        std::string typeName;
        if (container.aronType())
        {
            typeName = container.aronType()->getName();
            typeName = instance::sanitizeTypeName(typeName);
        }
        else
        {
            typeName = "(no Aron type)";
        }
        item->setText(int(Columns::TYPE), QString::fromStdString(typeName));
    }

}
